/*
===============================================================================

  FILE:  rangefloatcompressor.cpp
  
  CONTENTS:
  
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    Copyright (C) 2000  Martin Isenburg (isenburg@cs.unc.edu)
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    see header file
  
===============================================================================
*/

#include "rangefloatcompressor.h"

#include <stdlib.h>
#include <assert.h>

// 0  1  2  3  4   5   6   7    8    9  10     11    12  13  14  15  16  17  18   19   20   21    22    23    24

const I32 bitsLowTable[] = {
   0, 0, 1, 2, 3,  4,  5,  6,   7,   8,  9,    10,   11,  1,  2,  3,  4,  5,  6,   7,   8,   9,   10,   11,   12
};

const I32 bitsMaskTable[] = {
   0, 0, 1, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047,  1,  3,  7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4096
};

const I32 bitsHighTable[] = {
   0, 0, 0, 0, 0,  0,  0,  0,   0,   0,   0,    0,    0, 11, 11, 11, 11, 11, 11,  11,  11,  11,   11,   11,   11
};

RangeFloatCompressor::RangeFloatCompressor()
{
  bits = 0;
  bits_for_mantissa = (I32*)malloc(sizeof(I32)*256);
}

RangeFloatCompressor::~RangeFloatCompressor()
{
  free(bits_for_mantissa);
}

void RangeFloatCompressor::SetupCompressor(RangeEncoder* re, U32 contexts)
{
  assert(re);
  assert(contexts > 0);
  this->re = re;
  this->contexts = contexts;
  alloc_range_tables();
  reset_precision();
}

void RangeFloatCompressor::SetupDecompressor(RangeDecoder* rd, U32 contexts)
{
  assert(rd);
  assert(contexts > 0);
  this->rd = rd;
  this->contexts = contexts;
  alloc_range_tables();
  reset_precision();
}

void RangeFloatCompressor::alloc_range_tables()
{
  U32 c,i;
  
  rmSign = (RangeModel***)malloc(sizeof(RangeModel**)*contexts);
  rmExponent = (RangeModel***)malloc(sizeof(RangeModel**)*contexts);
  rmMantissaLow = (RangeModel***)malloc(sizeof(RangeModel**)*contexts);
  rmMantissaHigh = (RangeModel***)malloc(sizeof(RangeModel**)*contexts);

  for (c = 0; c < contexts; c++)
  {
    rmSign[c] = (RangeModel**)malloc(sizeof(RangeModel*)*256);
    rmExponent[c] = (RangeModel**)malloc(sizeof(RangeModel*)*256);
    rmMantissaLow[c] = (RangeModel**)malloc(sizeof(RangeModel*)*256);
    rmMantissaHigh[c] = (RangeModel**)malloc(sizeof(RangeModel*)*256);

    for (i = 0; i < 256; i++)
    {
      rmSign[c][i] = 0;
      rmExponent[c][i] = 0;
      rmMantissaLow[c][i] = 0;
      rmMantissaHigh[c][i] = 0; 
    }
  }

  rmZeroOrExtremeValue = (RangeModel**)malloc(sizeof(RangeModel*)*256);
  rmPositiveOrNegativeValue = (RangeModel**)malloc(sizeof(RangeModel*)*256);

  for (i = 0; i < 256; i++)
  {
    rmZeroOrExtremeValue[i] = 0;
    rmPositiveOrNegativeValue[i] = 0;
  }
}

void RangeFloatCompressor::dealloc_range_tables()
{
  U32 c,i;

  for (c = 0; c < contexts; c++)
  {
    for (i = 0; i < 256; i++)
    {
      if (rmSign[c][i]) delete rmSign[c][i];
      if (rmExponent[c][i]) delete rmExponent[c][i];
      if (rmMantissaLow[c][i]) delete rmMantissaLow[c][i];
      if (rmMantissaHigh[c][i]) delete rmMantissaHigh[c][i];
    }
    free(rmSign[c]);
    free(rmExponent[c]);
    free(rmMantissaLow[c]);
    free(rmMantissaHigh[c]);
  }
  free(rmSign);
  free(rmExponent);
  free(rmMantissaLow);
  free(rmMantissaHigh);
  
  for (i = 0; i < 256; i++)
  {
    if (rmZeroOrExtremeValue[i]) delete rmZeroOrExtremeValue[i];
    if (rmPositiveOrNegativeValue[i]) delete rmPositiveOrNegativeValue[i];
  }
  free(rmZeroOrExtremeValue);
  free(rmPositiveOrNegativeValue);
}

void RangeFloatCompressor::FinishCompressor()
{
  dealloc_range_tables();
}

void RangeFloatCompressor::FinishDecompressor()
{
  dealloc_range_tables();
}

void RangeFloatCompressor::compress_sign(I32 exponent, I32 sign, RangeEncoder* re_sign, RangeModel** rmSign)
{
  if (rmSign[exponent] == 0)
  {
    rmSign[exponent] = new RangeModel(2, 0, 1);
  }
  re_sign->encode(rmSign[exponent], sign);
//  fprintf(stderr,"compress_sign exp %d sign %d\n",exponent,sign);
}

I32 RangeFloatCompressor::decompress_sign(I32 exponent, RangeDecoder* rd_sign, RangeModel** rmSign)
{
  if (rmSign[exponent] == 0)
  {
    rmSign[exponent] = new RangeModel(2, 0, 0);
  }
  I32 sign = rd_sign->decode(rmSign[exponent]);
//  fprintf(stderr,"decompress_sign exp %d sign %d\n",exponent,sign);
  return sign;
}

void RangeFloatCompressor::compress_exponent(I32 exponentPred, I32 exponentReal, RangeEncoder* re_exponent, RangeModel** rmExponent)
{
  if (rmExponent[exponentPred] == 0)
  {
    rmExponent[exponentPred] = new RangeModel(256,0,1);
  }
  re_exponent->encode(rmExponent[exponentPred],exponentReal);
//  fprintf(stderr,"compress_exponent exp %d real %d\n",exponentPred,exponentReal);
}

I32 RangeFloatCompressor::decompress_exponent(I32 exponentPred, RangeDecoder* rd_exponent, RangeModel** rmExponent)
{
  if (rmExponent[exponentPred] == 0)
  {
    rmExponent[exponentPred] = new RangeModel(256,0,0);
  }

  I32 exponentReal = rd_exponent->decode(rmExponent[exponentPred]);
//  fprintf(stderr,"decompress_exponent exp %d real %d\n",exponentPred,exponentReal);
  return exponentReal;
}

I32 RangeFloatCompressor::compress_mantissa(I32 exponent, I32 mantissaPred, I32 mantissaReal, RangeEncoder* re_mantissa, RangeModel** rmMantissaHigh, RangeModel** rmMantissaLow)
{
  I32 c, sign;

  I32 bits_used = bits_for_mantissa[exponent];

  if (!bits_used) // all mantissa bits are discarded
  {
    return 0;
  }

  I32 bits_discarded = 23 - bits_used;

  I32 bits_mask = (1<<23)-(1<<bits_discarded);

  mantissaReal &= bits_mask;

  c = mantissaReal - (mantissaPred&bits_mask);

  if (c < (1-(1<<22)))
  {
    c += (1<<23);
  }
  else if (c > (1<<22))
  {
    c -= (1<<23); 
  }

  // now c should lie in the interval [ - (2^22 - 1)  ...  + (2^22) ]

  if (c == (1<<22))            // the extreme misprediction
  {
    sign = -2;
    c = 0;
  }
  else
  {
    c = c >> bits_discarded;  // discard the zero bits

    if (c > 0)                // positive
    {
      sign = 0;
    }
    else if (c < 0)            // negative
    {
      sign = 1;
      c = -c;
    }
    else                      // zero
    {
      sign = -1;
    }
  }

  I32 bitsLowUsed = bitsLowTable[bits_used];

  if (bitsLowUsed)
  {
    I32 bitsHighUsed = bitsHighTable[bits_used];

    if (rmMantissaLow[exponent] == 0)
    {
      rmMantissaLow[exponent] = new RangeModel((1<<bitsLowUsed),0,1);
      if (bitsHighUsed)
      {
        rmMantissaHigh[exponent] = new RangeModel((1<<bitsHighUsed),0,1);
      }
    }

    if (bitsHighUsed)
    {
      I32 bitsHigh = c >> bitsLowUsed;
      re_mantissa->encode(rmMantissaHigh[exponent], bitsHigh);
      I32 bitsLow = c & bitsMaskTable[bits_used];
      re_mantissa->encode(rmMantissaLow[exponent], bitsLow);
    }
    else
    {
      re_mantissa->encode(rmMantissaLow[exponent], c);
    }
  }

  if (sign < 0)      // zero or extreme
  {
    if (rmZeroOrExtremeValue[exponent] == 0)
    {
      rmZeroOrExtremeValue[exponent] = new RangeModel(2,0,1);
    }
    if (sign == -1) // zero
    {
      re_mantissa->encode(rmZeroOrExtremeValue[exponent], 0);
    }
    else            // extreme
    {
      re_mantissa->encode(rmZeroOrExtremeValue[exponent], 1);
    }
  }
  else              // positive or negative
  {
    if (rmPositiveOrNegativeValue[exponent] == 0)
    {
      rmPositiveOrNegativeValue[exponent] = new RangeModel(2,0,1);
    }
    if (sign == 0)  // positive
    {
      re_mantissa->encode(rmPositiveOrNegativeValue[exponent], 0);
    }
    else            // negative
    {
      re_mantissa->encode(rmPositiveOrNegativeValue[exponent], 1);
    }
  }
  return mantissaReal;
}

I32 RangeFloatCompressor::decompress_mantissa(I32 exponent, I32 mantissaPred, RangeDecoder* rd_mantissa, RangeModel** rmMantissaLow, RangeModel** rmMantissaHigh)
{
  I32 c = 0;

  I32 bits_used = bits_for_mantissa[exponent];

  if (!bits_used) // all mantissa bits are discarded
  {
    return 0;
  }

  I32 bitsLowUsed = bitsLowTable[bits_used];

  if (bitsLowUsed)
  {
    I32 bitsHighUsed = bitsHighTable[bits_used];

    if (rmMantissaLow[exponent] == 0)
    {
      rmMantissaLow[exponent] = new RangeModel((1<<bitsLowUsed),0,0);
      if (bitsHighUsed)
      {
        rmMantissaHigh[exponent] = new RangeModel((1<<bitsHighUsed),0,0);
      }
    }

    if (bitsHighUsed)
    {
      c = (rd_mantissa->decode(rmMantissaHigh[exponent]) << bitsLowUsed);
      c = c | rd_mantissa->decode(rmMantissaLow[exponent]);
    }
    else
    {
      c = rd_mantissa->decode(rmMantissaLow[exponent]);
    }
  }

  I32 bits_discarded = 23 - bits_used;

  if (c != 0) // decode sign
  {
    if (rmPositiveOrNegativeValue[exponent] == 0)
    {
      rmPositiveOrNegativeValue[exponent] = new RangeModel(2,0,0);
    }
    if (rd_mantissa->decode(rmPositiveOrNegativeValue[exponent]))
    {
      c = -(c << bits_discarded);
    }
    else
    {
      c = c << bits_discarded;
    }
  }
  else // decode zero or extreme
  {
    if (rmZeroOrExtremeValue[exponent] == 0)
    {
      rmZeroOrExtremeValue[exponent] = new RangeModel(2,0,0);
    }
    if (rd_mantissa->decode(rmZeroOrExtremeValue[exponent]))
    {
      c = (1<<(22));
    }
  }

  I32 bits_mask = (1<<23)-(1<<bits_discarded);

  c += (mantissaPred&bits_mask);

  // wrap back into valid range
  if (c < 0)
  {
    c += (1<<23);
  }
  else if (c >= (1<<23))
  {
    c -= (1<<23);
  }

  return c;
}

void RangeFloatCompressor::reset_precision()
{
  U32 i;
  range_exponent = -2;
  for (i = 0; i < 256; i++)
  {
    bits_for_mantissa[i] = 23;
  }
}

void RangeFloatCompressor::update_precision(F32 number)
{
  if (range_exponent == -2)
  {
    min_float = number;
    max_float = number;
    range_exponent = -1;
    return;
  }
  else
  {
    if (number < min_float)
    {
      min_float = number;
    }
    else if (number > max_float)
    {
      max_float = number;
    }
    else
    {
      return;
    }
  }
  U32F32 range;
  range.f32 = max_float - min_float;
  I32 exponent = (range.u32 >> 23) - ((range.u32 & 0x007FFFFF) == 0);
  if (exponent > range_exponent)
  {
    range_exponent = exponent;
    U32 c;
    I32 i,j;
    // the mantissa of numbers with exponent max_exponent will need to
    // preserve the (bits - (range_exponent - max_exponent)) highest
    // bits
    i = 22 - bits + range_exponent;
    if (i > 255) i = 255;
    j = 22;
    while (j >= 0 && i >= 0)
    {
      for (c = 0; c < contexts; c++)
      {
	if (rmMantissaLow[c][i]) {delete rmMantissaLow[c][i]; rmMantissaLow[c][i] = 0;};
	if (rmMantissaHigh[c][i]) {delete rmMantissaHigh[c][i]; rmMantissaHigh[c][i] = 0;};
      }
      bits_for_mantissa[i] = j;
      i--;
      j--;
    }
    while (i >= 0 && (bits_for_mantissa[i] != -1))
    {
      for (c = 0; c < contexts; c++)
      {
	if (rmMantissaLow[c][i]) {delete rmMantissaLow[c][i]; rmMantissaLow[c][i] = 0;};
	if (rmMantissaHigh[c][i]) {delete rmMantissaHigh[c][i]; rmMantissaHigh[c][i] = 0;};
      }
      bits_for_mantissa[i] = -1;
      i--;
    }
  }
}

F32 RangeFloatCompressor::Compress(F32 fPred, F32 fReal, U32 context)
{
  U32F32 uPred;
  U32F32 uReal;
  uPred.f32 = fPred;
  uReal.f32 = fReal;
  
  I32 signPred = (uPred.u32 & 0x80000000) == 0x80000000;
  I32 exponentPred = (uPred.u32 & 0x7F800000) >> 23;
  I32 mantissaPred = (uPred.u32 & 0x007FFFFF);

  I32 signReal = (uReal.u32 & 0x80000000) == 0x80000000;
  I32 exponentReal = (uReal.u32 & 0x7F800000) >> 23;
  I32 mantissaReal = (uReal.u32 & 0x007FFFFF);

  if (bits_for_mantissa[exponentReal] >= 0)
  {
    // compress exponent
    compress_exponent(exponentPred,exponentReal,re,rmExponent[context]);
    // compress sign
    if (signPred == signReal)
    {
      compress_sign(exponentReal,0,re,rmSign[context]); // correct predicted
    }
    else
    {
      compress_sign(exponentReal,1,re,rmSign[context]); // incorrect predicted
    }
    // compress mantissa
    if (signPred == signReal) // sign predicted correctly
    {
      if (exponentPred == exponentReal)
      {
        mantissaReal = compress_mantissa(exponentReal,mantissaPred,mantissaReal,re,rmMantissaLow[context],rmMantissaHigh[context]);
      }
      else if (exponentPred > exponentReal)
      {
        mantissaReal = compress_mantissa(exponentReal,0x007FFFFF,mantissaReal,re,rmMantissaLow[context],rmMantissaHigh[context]);
      }
      else
      {
        mantissaReal = compress_mantissa(exponentReal,0x0,mantissaReal,re,rmMantissaLow[context],rmMantissaHigh[context]);
      }
    }
    else if (signPred) // predicted negative, but in reality it is positive
    {
      mantissaReal = compress_mantissa(exponentReal,0x0,mantissaReal,re,rmMantissaLow[context],rmMantissaHigh[context]);
    }
    else // predicted positive, but in reality it is negative
    {
      mantissaReal = compress_mantissa(exponentReal,0x0,mantissaReal,re,rmMantissaLow[context],rmMantissaHigh[context]);
    }

    uReal.u32 = (uReal.u32 & 0xFF800000) | mantissaReal;

    // update precision
    if (exponentReal == 255)
    {
      // fprintf(stderr,"WARNING: exponentReal is 255 -> infinite number\n");
    }
    else
    {
      if (bits) update_precision(uReal.f32);
      else assert(uReal.f32 == fReal);
    }
    return uReal.f32;
  }
  else // the float is quantized to 0.0f
  {
    // compress only exponent
    compress_exponent(exponentPred,0,re,rmExponent[context]);
    return 0.0f;
  }
}

F32 RangeFloatCompressor::Decompress(F32 fPred, U32 context)
{
  U32F32 uPred;
  uPred.f32 = fPred;
  
  I32 exponentPred = (uPred.u32 & 0x7F800000) >> 23;
  I32 exponentReal = decompress_exponent(exponentPred,rd,rmExponent[context]);

  if (bits_for_mantissa[exponentReal] >= 0)
  {
    I32 signPred = (uPred.u32 & 0x80000000) == 0x80000000;
    I32 mantissaPred = (uPred.u32 & 0x007FFFFF);
    I32 signReal;
    I32 mantissaReal;
    // decompress sign
    if (decompress_sign(exponentReal,rd,rmSign[context]))
    {
      signReal = !signPred;
    }
    else
    {
      signReal = signPred;
    }
    // decompress mantissa
    if (signPred == signReal) // sign predicted correctly
    {
      if (exponentPred == exponentReal) // exponent predicted correctly
      {
        mantissaReal = decompress_mantissa(exponentReal,mantissaPred,rd,rmMantissaLow[context],rmMantissaHigh[context]);
      }
      else if (exponentPred > exponentReal) // overshot exponent prediction
      {
        mantissaReal = decompress_mantissa(exponentReal,0x007FFFFF,rd,rmMantissaLow[context],rmMantissaHigh[context]);
      }
      else // undershot exponent prediction
      {
        mantissaReal = decompress_mantissa(exponentReal,0x0,rd,rmMantissaLow[context],rmMantissaHigh[context]);
      }
    }
    else if (signPred) // predicted negative, but in reality it is positive
    {
      mantissaReal = decompress_mantissa(exponentReal,0x0,rd,rmMantissaLow[context],rmMantissaHigh[context]);
    }
    else // predicted positive, but in reality it is negative
    {
      mantissaReal = decompress_mantissa(exponentReal,0x0,rd,rmMantissaLow[context],rmMantissaHigh[context]);
    }
    // put together decompressed float value
    U32F32 uReal;
    if (signReal)
    {
      uReal.u32 = 0x80000000 | (exponentReal << 23) | mantissaReal;
    }
    else
    {
      uReal.u32 = (exponentReal << 23) | mantissaReal;
    }
    // update precision
    if (exponentReal == 255)
    {
      // fprintf(stderr,"WARNING: exponentReal is 255 -> infinite number\n");
    }
    else
    {
      if (bits) update_precision(uReal.f32);
    }
    return uReal.f32;
  }
  return 0.0f; // we should never get here
}
