/*
===============================================================================

  FILE:  svreader_geo.cpp

  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009  clement.courbet@ecp.fr

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svreader_geo.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>

//this is for compatibility with the new g++
#ifdef __GNUC__
	#include <ext/hash_map>
	namespace std { using namespace __gnu_cxx; }
	using namespace std;
#else
	#include <hash_map.h>
#endif

//#define PRINT_DEBUG_OUTPUT

#include "geo_structures.h"

using namespace Geo;

typedef hash_map<int, Geo::Point*> Point_hash;
typedef hash_map<int, Geo::Curve*> Curve_hash;
typedef hash_map<int, Geo::Surface*> Surface_hash;
typedef vector<Geo::Block*> Block_vector;

static Point_hash* points_hash = 0;
static Curve_hash* curves_hash = 0;
static Surface_hash* surfaces_hash = 0;
static Block_vector* blocks_vector = 0;
static Geo::StreamList verticesToStream;

bool SVreader_geo::open ( FILE* file )
{
	if ( file == 0 )
	{
		return false;
	}

	this->file = file;

	v_count = 0;
	c_count = 0;
	nverts=-1;
	ncells=-1;
	indexBase = 0;
	
	v_pnum_f = 3;
	c_pnum_f = 0;

	if(v_prop_f) delete[] v_prop_f;
	v_prop_f = new float[3];

	if(c_idx) delete[] c_idx;
	c_idx = new I32[8];
	if(c_final) delete[] c_final;
	c_final = new bool[8];
	for(int i=0; i<8; ++i) c_final[i]=false;

	if(next_idx) delete[] next_idx;
	next_idx = new I32[8];
	if(next_final) delete[] next_final;
	next_final = new bool[8];

	//explicit finalization stuff (not implemented)
	have_finalized = 0; 
	next_finalized = 0;
	if(finalized_vertices) delete[] finalized_vertices;
	finalized_vertices = new I32[8];

	//initialize static variables
	if (points_hash) delete points_hash;
	points_hash = 0;
	if (curves_hash) delete curves_hash;
	curves_hash = 0;
	if (surfaces_hash) delete surfaces_hash;
	surfaces_hash = 0;
	if (blocks_vector) delete blocks_vector;
	blocks_vector = 0;
	
	currentBlock = -1;
	verticesToStream.clear();

	if(!read_geometry())
	{
		return false;
	}	

	return true;
}

bool SVreader_geo::read_geometry ( )
{
	char line[1024];
	int index;
	int subdiv;
	int bytesRead;

	points_hash = new Point_hash();
	curves_hash = new Curve_hash();
	surfaces_hash = new Surface_hash();
	blocks_vector = new Block_vector();

	while(fgets(line, 1023, file))
	{
		if(sscanf(line, "Point ( %i )%n", &index, &bytesRead)==1)
		{
			//we found a point, read the coordinates
			Point *p = new Point(index, Point::TYPE_CORNER, 0); //each block this vertex belongs to will add a slot...
			float dummy;
			if(sscanf(line+bytesRead, " = { %f , %f , %f , %f } ;", &(p->pos[0]), &(p->pos[1]),&(p->pos[2]), &dummy)==4)
			{
#ifdef PRINT_DEBUG_OUTPUT
				fprintf(stderr, "Found a point with index %i and coords (%f, %f, %f)\n", index, p->pos[0], p->pos[1], p->pos[2]);
#endif
				points_hash->insert ( Point_hash::value_type ( index, p ) );
			}
			else
			{
				fprintf(stderr, "ERROR: Could not read coordinates for point %i\n", index);
				return false;
			}
		}	
		else if(sscanf(line, "Line ( %i )%n", &index, &bytesRead)==1)
		{
			//we found a straight line, read the control points
			int si, ei;
			if(sscanf(line+bytesRead, " = { %i , %i } ;", &si, &ei)==2)
			{
#ifdef PRINT_DEBUG_OUTPUT
				fprintf(stderr, "Found a straight line with index %i and points(%i, %i)\n", index, si, ei);
#endif
				Point_hash::iterator hit = points_hash->find ( si );
				if ( hit == points_hash->end() )
				{
					fprintf ( stderr,"ERROR: point %i is not in hash\n", si );
					exit ( 1 );
				}
				Point *sp = (*hit).second;
				hit = points_hash->find ( ei );
				if ( hit == points_hash->end() )
				{
					fprintf ( stderr,"ERROR: point %i is not in hash\n", ei );
					exit ( 1 );
				}
				Point *ep = (*hit).second;

				curves_hash->insert ( Curve_hash::value_type ( index, new Line(index, sp, ep) ) );
			}
			else
			{
				fprintf(stderr, "ERROR: Could not read start and end points for straight line %i\n", index);
				return false;
			}
		}
		else if(sscanf(line, "Circle ( %i )%n", &index, &bytesRead)==1)
		{
			//we found a circle line, read the control points
			int si, ei, ci;
			if(sscanf(line+bytesRead, " = { %i , %i , %i } ;", &si, &ci, &ei)==3)
			{
#ifdef PRINT_DEBUG_OUTPUT
				fprintf(stderr, "Found a circle with index %i and points(%i, %i, %i)\n", index, si, ci, ei);
#endif
				Point_hash::iterator hit = points_hash->find ( si );
				if ( hit == points_hash->end() )
				{
					fprintf ( stderr,"ERROR: point %i is not in hash\n", si );
					exit ( 1 );
				}
				Point *sp = (*hit).second;
				hit = points_hash->find ( ci );
				if ( hit == points_hash->end() )
				{
					fprintf ( stderr,"ERROR: point %i is not in hash\n", ci );
					exit ( 1 );
				}
				Point *cp = (*hit).second;
				hit = points_hash->find ( ei );
				if ( hit == points_hash->end() )
				{
					fprintf ( stderr,"ERROR: point %i is not in hash\n", ei );
					exit ( 1 );
				}
				Point *ep = (*hit).second;

				curves_hash->insert ( Curve_hash::value_type ( index, new Circle(index, sp, cp, ep) ) );
			}
			else
			{
				fprintf(stderr, "ERROR: Could not read control points for circle %i\n", index);
				return false;
			}
		}
		else if(sscanf(line, "Line Loop ( %i )%n", &index, &bytesRead)==1)
		{
			//we found a line loop, read the four control surfaces
			int li[4];
			if(sscanf(line+bytesRead, " = { %i , %i , %i , %i } ;", &li[0], &li[1], &li[2], &li[3])==4)
			{
#ifdef PRINT_DEBUG_OUTPUT
				fprintf(stderr, "Found a line loop with index %i and control curves(%i, %i, %i, %i)\n", index, li[0], li[1], li[2], li[3]);
#endif
				Curve ** lp = new Curve*[4];
				bool *reverse = new bool[4];
				for(int j=0; j<4; ++j)
				{
					if(li[j]<0) 
					{
						li[j] = -li[j];
						reverse[j] = true;
					}
					else
					{
						reverse[j] = false;
					}
					Curve_hash::iterator hit = curves_hash->find ( li[j] );
					if ( hit == curves_hash->end() )
					{
						fprintf ( stderr,"ERROR: point %i is not in hash\n", li[j] );
						exit ( 1 );
					}
					lp[j] = (*hit).second;
				}

				surfaces_hash->insert ( Surface_hash::value_type ( index, new Surface(index, lp, reverse) ) );
			}
			else
			{
				fprintf(stderr, "ERROR: Could not read control curves for line loop %i\n", index);
				return false;
			}
		}
		else if(sscanf(line, "Line Loop ( %i )%n", &index, &bytesRead)==1)
		{
			//we found a line loop, read the four control surfaces
			int li[4];
			if(sscanf(line+bytesRead, " = { %i , %i , %i , %i } ;", &li[0], &li[1], &li[2], &li[3])==4)
			{
#ifdef PRINT_DEBUG_OUTPUT
				fprintf(stderr, "Found a line loop with index %i and control curves(%i, %i, %i, %i)\n", index, li[0], li[1], li[2], li[3]);
#endif
				Curve ** lp = new Curve*[4];
				bool *reverse = new bool[4];
				for(int j=0; j<4; ++j)
				{
					if(li[j]<0) 
					{
						li[j] = -li[j];
						reverse[j] = true;
					}
					else
					{
						reverse[j] = false;
					}
					Curve_hash::iterator hit = curves_hash->find ( li[j] );
					if ( hit == curves_hash->end() )
					{
						fprintf ( stderr,"ERROR: curve %i is not in hash\n", li[j] );
						exit ( 1 );
					}
					lp[j] = (*hit).second;
				}

				surfaces_hash->insert ( Surface_hash::value_type ( index, new Surface(index, lp, reverse) ) );
			}
			else
			{
				fprintf(stderr, "ERROR: Could not read control curves for line loop %i\n", index);
				return false;
			}
		}
		else if(sscanf(line, "Surface Loop ( %i )%n", &index, &bytesRead)==1)
		{
			//we found a surface loop (block), read the six control surfaces
			int si[6];
			//NOTE: mind the face reordering and reorenting
			if(sscanf(line+bytesRead, " = { %i , %i , %i , %i , %i , %i } ;", &si[0], &si[1], &si[2], &si[5], &si[3], &si[4])==6)
			{
#ifdef PRINT_DEBUG_OUTPUT
				fprintf(stderr, "Found a surface loop with index %i and control surfaces (%i, %i, %i, %i, %i, %i)\n", index, si[0], si[1], si[2], si[3], si[4], si[5]);
#endif
				Surface * sp[6];
				for(int j=0; j<6; ++j)
				{
					if(si[j]<0) 
					{
						si[j] = -si[j];
					}
					Surface_hash::iterator hit = surfaces_hash->find ( si[j] );
					if ( hit == surfaces_hash->end() )
					{
						fprintf ( stderr,"ERROR: surface %i is not in hash\n", si[j] );
						exit ( 1 );
					}
					sp[j] = (*hit).second;
				}
				Block *block = new Block(index, sp);
				blocks_vector->push_back ( block );
			}
			else
			{
				fprintf(stderr, "ERROR: Could not read control curves for surface loop %i\n", index);
				return false;
			}
		}
		else if(sscanf(line, "Transfinite Line \"*\" = %i", &subdiv)==1)
		{
#ifdef PRINT_DEBUG_OUTPUT
			fprintf(stderr, "Defining all lines to transfinite (%i subdivisions)\n", subdiv);
#endif
			for(Curve_hash::iterator hit = curves_hash->begin(); hit != curves_hash->end (); ++hit )
			{
				(*hit).second->setSubdiv(subdiv);
			}
		}
		else if(sscanf(line, "Transfinite Line { %i %n", &index, &bytesRead)==1)
		{
#ifdef PRINT_DEBUG_OUTPUT
			fprintf(stderr, "Defining the following lines to transfinite:\n");
#endif
			//read all the lines
			int bytesR = 0;
			std::vector<Curve*> curves;
			do
			{
#ifdef PRINT_DEBUG_OUTPUT
			fprintf(stderr, " *%i\n", index);
#endif
				bytesRead+=bytesR;
				Curve_hash::iterator hit = curves_hash->find ( index );
				if ( hit == curves_hash->end() )
				{
					fprintf ( stderr,"ERROR: curve %i is not in hash\n", index );
					exit ( 1 );
				}
				curves.push_back((*hit).second);
			}
			while(sscanf(line+bytesRead, " , %i %n", &index, &bytesR)==1);
			//fprintf(stderr, "%s", line+bytesRead);
			//then read the subdivision level
			if(sscanf(line+bytesRead, "} = %i Using Progression 1 ;", &subdiv)==1)
			{
#ifdef PRINT_DEBUG_OUTPUT
				fprintf(stderr, "(%i subdivisions)\n", subdiv);
#endif
				for(std::vector<Curve*>::iterator vit = curves.begin(); vit != curves.end (); ++vit )
				{
					(*vit)->setSubdiv(subdiv);
				}
			}
			else
			{
				fprintf(stderr, "ERROR: Could not read subdivision level\n");
				return false;
			}
		}
	}

	return true;
}

void SVreader_geo::close ( bool close_file )
{
	// close of SVreader_svg
	if ( close_file && file && file != stdin ) fclose ( file );
	file = 0;
	have_finalized = 0; next_finalized = 0;
	delete [] finalized_vertices;
	finalized_vertices = 0;

	if(points_hash) delete points_hash;
	points_hash = NULL;
	if(curves_hash) delete curves_hash;
	curves_hash = NULL;
	if(surfaces_hash) delete surfaces_hash;
	surfaces_hash = NULL;
	if(blocks_vector) delete blocks_vector;
	blocks_vector = NULL;

	// close of SVreader interface
	v_count = -1;
	c_count = -1;
	if ( v_prop_f ) delete [] v_prop_f;
	v_prop_f = 0;
	if ( c_final ) delete [] c_final;
	c_final = 0;
	if ( c_idx ) delete [] c_idx;
	c_idx = 0;
	if ( next_idx ) delete [] next_idx;
	next_idx = NULL;
	if ( next_final ) delete [] next_final;
	next_final = NULL;
}

SVevent SVreader_geo::read_element()
{
	have_finalized = next_finalized = 0;
	
	if(currentBlock==-1)
	{
		//fetch the initial block
		currentBlock=0;
		if(currentBlock<(int)blocks_vector->size())
		{
			(*blocks_vector)[currentBlock]->beginMeshing();
			(*blocks_vector)[currentBlock]->getNextHex(next_idx, verticesToStream, next_final, indexBase);
		}
		else
			return SV_EOF;
	}
	else if (currentBlock >= (int)blocks_vector->size()) //means that we are done
	{
		return SV_EOF;
	}

	//if there are still unstreamed vertices, stream them
	if(!verticesToStream.empty())
	{
		v_idx = verticesToStream.front()->first;
		memcpy(v_prop_f, verticesToStream.front()->second, 3*sizeof(float));
#ifdef PRINT_DEBUG_OUTPUT
		fprintf(stderr, "streaming vertex %i\n", v_idx);
#endif
		delete[] verticesToStream.front()->second;
		delete verticesToStream.front();
		verticesToStream.pop_front();
		++v_count;
		return SV_VERTEX;
	}
	//no vertices are left to stream, output the current hex
#ifdef PRINT_DEBUG_OUTPUT
	fprintf(stderr, "streaming hex ");
#endif
	for(int i=0; i<8; ++i)
	{
		c_final[i] = next_final[i];
		c_idx[i] = next_idx[i];
#ifdef PRINT_DEBUG_OUTPUT
		fprintf(stderr, "%i ", c_idx[i]);
#endif
	}
	++c_count;
#ifdef PRINT_DEBUG_OUTPUT
	fprintf(stderr, "\n\n");
#endif
	//and fetch a new one
	if(!(*blocks_vector)[currentBlock]->getNextHex(next_idx, verticesToStream, next_final, indexBase))
	{
#ifdef PRINT_DEBUG_OUTPUT
		fprintf(stderr, "fetching new block\n");
#endif
		(*blocks_vector)[currentBlock]->endMeshing();
		//last hex in block, so get a new block, so get a new block.
		if( ++currentBlock < (int)blocks_vector->size())
		{
			(*blocks_vector)[currentBlock]->beginMeshing();
			(*blocks_vector)[currentBlock]->getNextHex(next_idx, verticesToStream, next_final, indexBase);
		}
	}

	return SV_HEXAHEDRON;
}

SVevent SVreader_geo::read_event()
{
	if ( have_finalized )
	{
		final_idx = finalized_vertices[next_finalized];
		have_finalized--; next_finalized++;
		return SV_FINALIZED;
	}
	else
	{
		return read_element();
	}
}

SVreader_geo::SVreader_geo()
{
	// init of SVreader_geo
	file = 0;
	have_finalized = 0; next_finalized = 0;
	finalized_vertices = NULL;
	nverts=-1;
	ncells=-1;
	points_hash=NULL;
	curves_hash=NULL;
	surfaces_hash=NULL;
	blocks_vector=NULL;
	next_idx = NULL;
	next_final = NULL;
}

SVreader_geo::~SVreader_geo()
{
	// clean-up for SVreader_geo
}
