/*
===============================================================================

  FILE:  svreader_mesh3d.cpp
  
  CONTENTS:
  
    see corresponding header file
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2008  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    see corresponding header file
  
===============================================================================
*/
#include "svreader_mesh3d.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>

#include "vecnfv.h"

#define MAX_LINE_SIZE 1024

bool SVreader_mesh3d::open(FILE* file)
{
  if (file == 0)
  {
    return false;
  }

#ifdef _WIN32
  if (file == stdin)
  {
    if(_setmode( _fileno( stdin ), _O_TEXT ) == -1 )
    {
      fprintf(stderr, "ERROR: cannot set stdin to text (translated) mode\n");
    }
  }
#endif

  this->file = file;

  skipped_lines = 0;
  line = (char*)malloc(sizeof(char)*MAX_LINE_SIZE);
  if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
  {
    free(line);
    line = 0;
    return false;
  }

  type = SV_TET;  // only TET meshes are supported
  c_nverts = 4;
  post_order = false;

  if (strstr(line,"NETGEN"))
  {
    if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
    {
      free(line);
      line = 0;
      return false;
    }
  }

  char dummy[MAX_LINE_SIZE];
  // look for header information
  if (sscanf(&(line[0]), "%d %s", &nverts, dummy) == 1)
  {
    fprintf(stderr, "nverts: %d\n", nverts);
  }
  else
  {
    fprintf(stderr, "ERROR: corrupt Mesh3D file\n");
    return false;
  }

  // read next line
  if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
  {
    free(line);
    line = 0;
    return false;
  }

  // how many vertex properties (all float)
  v_pnum_f = sscanf(&(line[0]), "%s %s %s %s %s %s %s %s %s %s %s %s", dummy, dummy, dummy, dummy, dummy, dummy, dummy, dummy, dummy, dummy, dummy, dummy);
  if (v_prop_f) delete [] v_prop_f;
  v_prop_f = new float[v_pnum_f];
  if (v_pnum_f >= 3 && v_pnum_f <= 8)
  {
    fprintf(stderr, "v_pnum_f: %d\n", v_pnum_f);
  }
  else if (v_pnum_f > 8)
  {
    fprintf(stderr,"WARNING: forcing property count from %d to 8\n",v_pnum_f);
    v_pnum_f = 8;
  }
  else
  {
    fprintf(stderr,"WARNING: forcing property count from %d to 3\n",v_pnum_f);
    v_pnum_f = 3;
  }

  v_count = 0;
  c_count = 0;

  c_nverts = 4;

  c_idx = new I32[4];   // only TET meshes are supported
  c_final = new bool[4];

  c_final[0] = false; 
  c_final[1] = false;
  c_final[2] = false;
  c_final[3] = false;

  return true;
}

void SVreader_mesh3d::close(bool close_file)
{
  // close of SVreader_mesh3d
  if (skipped_lines) fprintf(stderr,"WARNING: skipped %d lines.\n",skipped_lines);
  skipped_lines = 0;
  if (close_file && file && file != stdin) fclose(file);
  file = 0;
  if (line)
  {
    free(line);
    line = 0;
  }
  // close of SVreader interface
  v_count = -1;
  c_count = -1;
  if (v_prop_f) delete [] v_prop_f;
  v_prop_f = 0;
  if (c_idx) delete [] c_idx;
  c_idx = 0;
  if (c_final) delete [] c_final;
  c_final = 0; 
}

SVevent SVreader_mesh3d::read_element()
{
  if (line == 0)
  {
    return SV_EOF;
  }

  if (v_count < nverts)
  {
    v_idx = v_count;
    if (v_pnum_f == 3)
    {
      while(sscanf(&(line[0]), "%f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2])) != 3)
      {
        skipped_lines++;
        if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
        {
          if (v_count < nverts || c_count < ncells)
          {
            fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
            fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
          }
          free(line);
          line = 0;
          return SV_EOF;
        }
      }
    }
    else if (v_pnum_f == 4)
    {
      while(sscanf(&(line[0]), "%f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3])) != 4)
      {
        skipped_lines++;
        if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
        {
          if (v_count < nverts || c_count < ncells)
          {
            fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
            fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
          }
          free(line);
          line = 0;
          return SV_EOF;
        }
      }
    }
    else if (v_pnum_f == 5)
    {
      while(sscanf(&(line[0]), "%f %f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]), &(v_prop_f[4])) != 5)
      {
        skipped_lines++;
        if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
        {
          if (v_count < nverts || c_count < ncells)
          {
            fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
            fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
          }
          free(line);
          line = 0;
          return SV_EOF;
        }
      }
    }
    else if (v_pnum_f == 6)
    {
      while(sscanf(&(line[0]), "%f %f %f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]), &(v_prop_f[4]), &(v_prop_f[5])) != 6)
      {
        skipped_lines++;
        if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
        {
          if (v_count < nverts || c_count < ncells)
          {
            fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
            fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
          }
          free(line);
          line = 0;
          return SV_EOF;
        }
      }
    }
    else if (v_pnum_f == 7)
    {
      while(sscanf(&(line[0]), "%f %f %f %f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]), &(v_prop_f[4]), &(v_prop_f[5]), &(v_prop_f[6])) != 7)
      {
        skipped_lines++;
        if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
        {
          if (v_count < nverts || c_count < ncells)
          {
            fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
            fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
          }
          free(line);
          line = 0;
          return SV_EOF;
        }
      }
    }
    else if (v_pnum_f == 8)
    {
      while(sscanf(&(line[0]), "%f %f %f %f %f %f %f %f", &(v_prop_f[0]), &(v_prop_f[1]), &(v_prop_f[2]), &(v_prop_f[3]), &(v_prop_f[4]), &(v_prop_f[5]), &(v_prop_f[6]), &(v_prop_f[7])) != 8)
      {
        skipped_lines++;
        if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
        {
          if (v_count < nverts || c_count < ncells)
          {
            fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
            fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
          }
          free(line);
          line = 0;
          return SV_EOF;
        }
      }
    }
    else
    {
      fprintf(stderr,"FATAL ERROR: vertex property count of %d not implemented\n",v_pnum_f);
      exit(0);
    }
    v_count++;
    if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
    {
      if (v_count < nverts || c_count < ncells)
      {
        fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
        fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
      }
      free(line);
      line = 0;
    }
    return SV_VERTEX;
  }
  else if (c_count < ncells || ncells == -1)
  {
    if (ncells == -1)
    {
      char dummy[MAX_LINE_SIZE];
      // look for header information
      if (sscanf(&(line[0]), "%d %s", &ncells, dummy) == 1)
      {
        fprintf(stderr, "ncells: %d\n", ncells);
      }
      else
      {
        fprintf(stderr, "ERROR: corrupt Mesh3D file\n");
        return SV_ERROR;
      }
      // read next line
      if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
      {
        free(line);
        line = 0;
        if (c_count < ncells)
        {
          fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
        }
        return SV_EOF;
      }
    }
    int mat;
    while (sscanf(&(line[0]), "%d %d %d %d %d", &(mat), &(c_idx[0]), &(c_idx[1]), &(c_idx[2]), &(c_idx[3])) != 5)
    {
      skipped_lines++;
      if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
      {
        if (c_count < ncells)
        {
          fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
        }
        free(line);
        line = 0;
        return SV_EOF;
      }
    }
    c_idx[0]--; c_idx[1]--;  c_idx[2]--; c_idx[3]--;
    c_count++;
    if (fgets(line, sizeof(char) * MAX_LINE_SIZE, file) == 0)
    {
      if (c_count < ncells)
      {
        fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
      }
      free(line);
      line = 0;
    }
    return SV_TETRAHEDRON;
  }
  free(line);
  line = 0;
  return SV_EOF;
}

SVevent SVreader_mesh3d::read_event()
{
  return read_element();
}

SVreader_mesh3d::SVreader_mesh3d()
{
  // init of SVreader_mesh3d
  file = 0;
  skipped_lines = 0;
  line = 0;
}

SVreader_mesh3d::~SVreader_mesh3d()
{
  // clean-up for SVreader_mesh3d
}
