/*
===============================================================================

  FILE:  SVreader_svg.cpp

  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009  clement.courbet@ecp.fr

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svreader_svg.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>

#if !(defined(i386) || defined(_WIN32) || defined(AMD64))
#define BIG_ENDIAN_SYSTEM
#endif

//#define PRINT_DEBUG_OUTPUT

#define SVG_VERSION 1

typedef enum {
  SVG_VERTEX = 0,
  SVG_TETRAHEDRON = 1,
  SVG_HEXAHEDRON = 2
} SVElementType;

//#define USE_ASM_BYTESWAP
#ifdef USE_ASM_BYTESWAP
extern inline void swap_endianness ( U32 *x );
#else
inline void swap_endianness ( U32 *x )
{
  *x = ( (*x)>>24 ) |
      ( ( (*x)<<8 ) & 0x00FF0000 ) |
      ( ( (*x)>>8 ) & 0x0000FF00 ) |
      ( (*x)<<24 );
}
#endif

inline void SVreader_svg::readInt32 ( I32 *i )
{
  fread ( i, 4, 1, file );
#ifdef BIG_ENDIAN_SYSTEM
  swap_endianness ( (U32*)i );
#endif
#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr, "read I32\t%i\n", *i);
#endif
}

inline void SVreader_svg::readFloat ( float *f )
{
  fread ( f, 4, 1, file );
#ifdef BIG_ENDIAN_SYSTEM
  swap_endianness ( (U32*)f );
#endif
#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr, "read float\t%g\n", *f);
#endif
}

inline void SVreader_svg::readNFloats ( float *f, int n )
{
  fread ( f, 4, n, file );
#ifdef BIG_ENDIAN_SYSTEM
  float *fp = f;
  for(int i=0; i<n; ++i, ++fp)
    swap_endianness ( (U32*)fp );
#endif
#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr, "read %i floats\n", n);
#endif
}

inline void SVreader_svg::readUChar ( unsigned char *c )
{
  fread ( c, 1, 1, file );
#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr, "read uint8_t\t%i\n", (int)*c);
#endif
}

bool SVreader_svg::open ( FILE* file )
{
  if ( file == 0 )
  {
    return false;
  }

#ifdef _WIN32
  if ( file == stdin )
  {
    if ( _setmode ( _fileno ( stdin ), _O_BINARY ) == -1 )
    {
      fprintf ( stderr, "ERROR: cannot set stdin to binary (untranslated) mode\n" );
    }
  }
#endif

  this->file = file;

  unsigned char version;
  readUChar(&version);
  // read version
  if ( version != SVG_VERSION )
  {
    fprintf ( stderr,"ERROR: wrong SVreader (need %d but this is SVreader_svg %d)\n",version,SVG_VERSION );
    if ( version == EOF ) fprintf ( stderr,"       check ... maybe the file (or the stream) is empty\n" );
    exit ( 0 );
  }

  read_header();

  v_count = 0;
  c_count = 0;

  if ( v_pnum_f ) v_prop_f = new float[v_pnum_f];
  if ( c_pnum_f ) c_prop_f = new float[c_pnum_f];

  c_idx = new I32[8];
  c_final = new bool[8];
  finalized_vertices = new I32[8];
  return true;
}

void SVreader_svg::close ( bool close_file )
{
  // close of SVreader_svg
  if ( close_file && file && file != stdin ) fclose ( file );
  file = 0;
  have_finalized = 0; next_finalized = 0;
  delete [] finalized_vertices;
  finalized_vertices = 0;

  // close of SVreader interface
  v_count = -1;
  c_count = -1;
  if ( v_prop_f ) delete [] v_prop_f;
  v_prop_f = 0;
  if ( c_idx ) delete [] c_idx;
  c_idx = 0;
  if ( c_final ) delete [] c_final;
  c_final = 0;
}

SVevent SVreader_svg::read_element()
{
  
  have_finalized = next_finalized = 0;
  //read element type
  if( (v_count<nverts) || (c_count<ncells))
  {
    unsigned char elementType;
    readUChar(&elementType);
    if ( elementType == SVG_VERTEX ) // next element is a vertex
    {
      if ( v_pnum_f )
      {
        readNFloats ( v_prop_f, v_pnum_f );
      }
      v_idx = v_count;
      v_count++;
      return SV_VERTEX;
    }
    else if ( elementType == SVG_TETRAHEDRON ) // next element is a tetrahedron
    {
      c_count++;
      for ( int i = 0; i < 4; i++ )
      {
        readInt32(&(c_idx[i])); //read the index
        if ( c_idx[i] < 0 )
        {
          c_idx[i] = v_count+c_idx[i];
          c_final[i] = true;
          finalized_vertices[have_finalized] = c_idx[i];
          have_finalized++;
        }
        else
        {
          c_idx[i] = c_idx[i]-1;
          c_final[i] = false;
        }
      }
      //read cell data
      if ( c_pnum_f )
      {
        readNFloats ( c_prop_f, c_pnum_f );
      }
      c_nverts = 4;
      return SV_TETRAHEDRON;
    }
    else if ( elementType == SVG_HEXAHEDRON ) // next element is a hexahedron
    {
      c_count++;
      for ( int i = 0; i < 8; i++ )
      {
        readInt32(&(c_idx[i])); //read the index
        if ( c_idx[i] < 0 )
        {
          c_idx[i] = v_count+c_idx[i];
          c_final[i] = true;
          finalized_vertices[have_finalized] = c_idx[i];
          have_finalized++;
        }
        else
        {
          c_idx[i] = c_idx[i]-1;
          c_final[i] = false;
        }
      }
      //read cell data
      if ( c_pnum_f )
      {
        readNFloats ( c_prop_f, c_pnum_f );
        //fprintf(stderr, "read %f\n", *c_prop_f);
      }
      c_nverts = 8;
      return SV_HEXAHEDRON;
    }
    else
    {
      printf("ERROR: SVreader_svg: unsupported element type %i (v_count=%d/%d, c_count=%d/%d)\n", elementType, v_count, nverts, c_count, ncells);
      exit(1);
    }
  }

  if ( nverts != -1 && v_count != nverts )
  {
    fprintf ( stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts );
  }
  nverts = v_count;
  if ( ncells != -1 && c_count != ncells )
  {
    fprintf ( stderr,"WARNING: wrong face count: c_count (%d) != ncells (%d)\n", c_count, ncells );
  }
  ncells = c_count;
  return SV_EOF;
}

SVevent SVreader_svg::read_event()
{
  if ( have_finalized )
  {
    final_idx = finalized_vertices[next_finalized];
    have_finalized--; next_finalized++;
    return SV_FINALIZED;
  }
  else
  {
    return read_element();
  }
}

void SVreader_svg::read_header()
{
  // read nverts
  readInt32(&nverts);
  // read ncells
  readInt32(&ncells);
  // read how many floating-point attributes per vertex
  readInt32(&v_pnum_f);
  // read how many floating-point attributes per cell
  readInt32(&c_pnum_f);

  // read bounding box of floating-point attributes (per vertex)
  if ( v_pnum_f )
  {
    if ( v_pmin_f ) delete [] v_pmin_f;
    if ( v_pmax_f ) delete [] v_pmax_f;
    v_pmin_f = new float[v_pnum_f];
    v_pmax_f = new float[v_pnum_f];
    readNFloats ( v_pmin_f, v_pnum_f );
    readNFloats ( v_pmax_f, v_pnum_f );
  }
  // read bounding box of floating-point attributes (per cell)
  if ( c_pnum_f )
  {
    if ( c_pmin_f ) delete [] c_pmin_f;
    if ( c_pmax_f ) delete [] c_pmax_f;
    c_pmin_f = new float[c_pnum_f];
    c_pmax_f = new float[c_pnum_f];
    readNFloats ( c_pmin_f, c_pnum_f );
    readNFloats ( c_pmax_f, c_pnum_f );
  }

  //reserved data
  I32 dummy;
  readInt32(&dummy);
  if(dummy!=0)
  {
    fprintf(stderr, "Error: SVReader_svg: reserved bytes must be zero !\n");
    exit(1);
  }
  readInt32(&dummy);
  if(dummy!=0)
  {
    fprintf(stderr, "Error: SVReader_svg: reserved bytes must be zero !\n");
    exit(1);
  }
}

SVreader_svg::SVreader_svg()
{
  // init of SVreader_svg
  file = 0;
  have_finalized = 0; next_finalized = 0;
}

SVreader_svg::~SVreader_svg()
{
  // clean-up for SVreader_svg
}
