/*
===============================================================================

  FILE:  sv2sv.cpp
  
  CONTENTS:
  
    This (experimental) program converts streaming volume meshes to and from
    the specified format.

  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2005-07 martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    15 August 2008 -- added hexahedra support after missing Henna all night long
    21 April 2008 -- added option to flip orientation of the tetrahedra 
    12 October 2007 -- using svreadopener and svwriteopener for file IO 
    25 May 2005 -- now can convert between OFF, TET, SVA, SVB, and SVC
    21 February 2005 -- created after a yummy backed Feta meal in the Mensa
  
===============================================================================
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "vec3fv.h"

#include "svreadopener.h"
#include "svwriteopener.h"

#include "svreadpreascompactpre.h"

#include "svwritebuffered.h"

static void usage()
{
  fprintf(stderr,"usage:\n");
  fprintf(stderr,"sv2sv -compact -precomp_bb -i mesh.off -o mesh.svc \n");
  fprintf(stderr,"sv2sv -i mesh.sva -o mesh.svc -delay 0\n");
  fprintf(stderr,"sv2sv -i mesh.sva -o mesh.svc -bits 12 -delay 100\n");
  fprintf(stderr,"sv2sv -i mesh.sva -o mesh.svc -bb -3.4 -2.2 -4.1 2.3 1.7 4.3\n");
  fprintf(stderr,"sv2sv -i mesh.svb.gz -o mesh.svc -cbm\n");
  fprintf(stderr,"sv2sv -i mesh.svc -osva | more\n");
  fprintf(stderr,"sv2sv -h\n");
  exit(1);
}

static bool positive_orientation(float** vertices)
{
  float u1[3], u2[3], u3[3], u1u2[3];
  VecSubtract3fv(u1, vertices[1], vertices[0]);
  VecSubtract3fv(u2, vertices[2], vertices[0]);
  VecSubtract3fv(u3, vertices[3], vertices[0]);
  VecCrossProd3fv(u1u2, u1, u2);
  return (VecDotProd3fv(u1u2, u3) >= 0.0f);
}

int main(int argc, char *argv[])
{
  int i;
  bool compact = false;
  bool flip = false;
  bool reorient = false;
  bool unfinalize = false;
  bool compute_bb = false;
  int set_v_pnum_f = 0;
  float* set_v_pmin_f = 0;
  float* set_v_pmax_f = 0;
  clock_t clock_start, clock_finish;
  time_t time_start, time_finish;

  SVreader* svreader = 0;
  SVreadDereferenced* svreaddereferenced = 0;
  SVreadOpener* svreadopener = 0;

  SVwriter* svwriter = 0;
  SVwriteOpener* svwriteopener = 0;

  // first look over the command line arguments for input and output
  for (i = 1; i < argc; i++)
  {
    if (strncmp(argv[i],"-i",2) == 0)
    {
      if (svreadopener == 0)
      {
        // apparently there is a point reader to be opened
        svreadopener = new SVreadOpener();
      }
      if (strcmp(argv[i],"-i") == 0) // the commandline '-i' followed by the input filename
      {
        i++;
        svreadopener->set_file_name(argv[i]);
      }
      else // the commandline '-ixxx' specifies the input format 'xxx'
      {
        svreadopener->set_file_format(&(argv[i][2]));
      }
    }
    else if (strncmp(argv[i],"-o",2) == 0)
    {
      if (svwriteopener == 0)
      {
        // apparently there is a point writer to be opened
        svwriteopener = new SVwriteOpener();
      }
      if (strcmp(argv[i],"-o") == 0) // the commandline '-o' followed by the output filename
      {
        i++;
        svwriteopener->set_file_name(argv[i]);
      }
      else // the commandline '-oxxx' specifies the output format 'xxx'
      {
        svwriteopener->set_file_format(&(argv[i][2]));
      }
    }
  }
  // second look over the command line arguments for options
  for (i = 1; i < argc; i++)
  {
    if ((strncmp(argv[i],"-i",2) == 0) || (strncmp(argv[i],"-o",2) == 0))
    {
      if ((strcmp(argv[i],"-i") == 0) || (strcmp(argv[i],"-o") == 0)) i++;
      // already processed
    }
    else if (strcmp(argv[i],"-b") == 0 || strcmp(argv[i],"-bits") == 0)
    {
      i++;
      if (svwriteopener) svwriteopener->set_compression_bits(atoi(argv[i]));
      else fprintf(stderr,"WARNING: no output specified. ignoring '%s %s'\n", argv[i-1], argv[i]);
    }
    else if (strcmp(argv[i],"-b") == 0 || strcmp(argv[i],"-cellbits") == 0)
    {
      i++;
      if (svwriteopener) svwriteopener->set_celldata_compression_bits(atoi(argv[i]));
      else fprintf(stderr,"WARNING: no output specified. ignoring '%s %s'\n", argv[i-1], argv[i]);
    }
    else if (strcmp(argv[i],"-delay") == 0)
    {
      i++;
      if (svwriteopener) svwriteopener->set_maximum_delay(atoi(argv[i]));
      else fprintf(stderr,"WARNING: no output specified. ignoring '%s %s'\n", argv[i-1], argv[i]);
    }
    else if (strcmp(argv[i],"-dry") == 0)
    {
      if (svwriteopener) svwriteopener->set_dry_run();
      else fprintf(stderr,"WARNING: no output specified. ignoring '%s'\n", argv[i]);
    }
    else if (strcmp(argv[i],"-cbm") == 0)
    {
      if (svwriteopener) svwriteopener->set_cbm_precision();
      else fprintf(stderr,"WARNING: no output specified. ignoring '%s'\n", argv[i]);
    }
    else if (strcmp(argv[i],"-compact") == 0 || strcmp(argv[i],"-vcompact") == 0)
    {
      compact = true;
    }
    else if (strcmp(argv[i],"-unfinal") == 0 || strcmp(argv[i],"-unfinalize") == 0)
    {
      unfinalize = true;
    }
    else if (strcmp(argv[i],"-flip") == 0)
    {
      flip = true;
    }
    else if (strcmp(argv[i],"-reorient") == 0)
    {
      flip = true;
      reorient = true;
    }
    else if (strcmp(argv[i],"-compute_bb") == 0 || strcmp(argv[i],"-precomp_bb") == 0)
    {
      compute_bb = true;
    }
    else if (strcmp(argv[i],"-v_pnum_f") == 0)
    {
      i++;
      set_v_pnum_f  = atoi(argv[i]);
    }
    else if (strcmp(argv[i],"-v_pminmax_f") == 0 || strcmp(argv[i],"-bb") == 0)
    {
      int j;
      if (set_v_pnum_f)
      {
        set_v_pmin_f = new float[set_v_pnum_f];
        set_v_pmax_f = new float[set_v_pnum_f];
        if (i + 2*set_v_pnum_f >= argc) usage();
        for (j = 0; j < set_v_pnum_f; j++) set_v_pmin_f[j] = (float)atof(argv[++i]);
        for (j = 0; j < set_v_pnum_f; j++) set_v_pmax_f[j] = (float)atof(argv[++i]);
      }
      else
      {
        set_v_pmin_f = new float[3];
        set_v_pmax_f = new float[3];
        if (i + 6 >= argc) usage();
        for (j = 0; j < 3; j++) set_v_pmin_f[j] = (float)atof(argv[++i]);
        for (j = 0; j < 3; j++) set_v_pmax_f[j] = (float)atof(argv[++i]);
      }
    }
    else if (strcmp(argv[i],"-high") == 0)
    {
      i++;
      if (svwriteopener) svwriteopener->set_high_bits(atoi(argv[i]));
      else fprintf(stderr,"WARNING: no output specified. ignoring '%s %s'\n", argv[i-1], argv[i]);
    }
    else if (strcmp(argv[i],"-cellhigh") == 0)
    {
      i++;
      if (svwriteopener) svwriteopener->set_celldata_high_bits(atoi(argv[i]));
      else fprintf(stderr,"WARNING: no output specified. ignoring '%s %s'\n", argv[i-1], argv[i]);
    }
    else if ((i == argc-2) && svreadopener == 0 && svwriteopener == 0)
    {
      // second to last argument unknown and neither input nor output specified yet ... try it
      svreadopener = new SVreadOpener();
      svreadopener->set_file_name(argv[i]);
    }
    else if ((i == argc-1) && svreadopener == 0)
    {
      // last argument unknown and no input specified yet ... try it
      svreadopener = new SVreadOpener();
      svreadopener->set_file_name(argv[i]);
    }
    else if ((i == argc-1) && svwriteopener == 0)
    {
      // last argument unknown and no output specified yet ... try it
      svwriteopener = new SVwriteOpener();
      svwriteopener->set_file_name(argv[i]);
    }
    else
    {
      usage();
    }
  }
  
  // make sure we have a readopener
  if (svreadopener == 0)
  {
    fprintf(stderr,"ERROR: there is no input\n");
    usage();
  }

  // open the mesh reader
  if (!(svreader = svreadopener->open()))
  {
    usage();
  }

  if (set_v_pnum_f) svreader->v_pnum_f = set_v_pnum_f;
  if (set_v_pmin_f && set_v_pmax_f) svwriter->set_v_pbox_f(set_v_pmin_f, set_v_pmax_f);

  if (compute_bb)
  {
    if (svreader->compute_bounding_box())
    {
      svreadopener->reopen(svreader);
      if (set_v_pnum_f) svreader->v_pnum_f = set_v_pnum_f;
      if (set_v_pmin_f && set_v_pmax_f) svwriter->set_v_pbox_f(set_v_pmin_f, set_v_pmax_f);
    }
  }

  // if we have a mesh writer then open the file
  if (svwriteopener)
  {
    if (!(svwriter = svwriteopener->open()))
    {
      usage();
    }
  }

  SVevent event;

  // start clock
  time_start = time(0);
  clock_start = clock();

  if (compact)
  {
    fprintf(stderr,"WARNING: using 'readPreAsCompactPre' ...\n");
    SVreadPreAsCompactPre* svreadpreascompactpre = new SVreadPreAsCompactPre();
    svreadpreascompactpre->open(svreader);
    svreaddereferenced = svreadpreascompactpre;
    svreader = svreadpreascompactpre;
  }
  else if (reorient)
  {
    fprintf(stderr,"WARNING: using 'readDereferenced' ...\n");
    svreaddereferenced = new SVreadDereferenced();
    svreaddereferenced->open(svreader);
    svreader = svreaddereferenced;
  }

  if (svwriter)
  {
    if (svreader->nverts != -1) svwriter->set_nverts(svreader->nverts);
    if (svreader->ncells != -1) svwriter->set_ncells(svreader->ncells);
    if (svreader->v_pnum_f) svwriter->set_v_pnum_f(svreader->v_pnum_f);
    if (svreader->v_pmin_f && svreader->v_pmax_f) svwriter->set_v_pbox_f(svreader->v_pmin_f, svreader->v_pmax_f);
    if (svreader->c_pnum_f) svwriter->set_c_pnum_f(svreader->c_pnum_f);
    if (svreader->c_pmin_f && svreader->c_pmax_f) svwriter->set_c_pbox_f(svreader->c_pmin_f, svreader->c_pmax_f);

    if (unfinalize)
    {
      bool c_final[] = {false,false,false,false,false,false,false,false};
      while ((event = svreader->read_element())!=SV_EOF)
      {
        switch (event)
        {
        case SV_VERTEX:
          svwriter->write_vertex(svreader->v_prop_f);
          break;
        case SV_TETRAHEDRON:
          if (flip)
          {
            if (!reorient || !positive_orientation(svreaddereferenced->c_v_prop_f))
            {
              I32 h = svreader->c_idx[2];
              svreader->c_idx[2] = svreader->c_idx[3];
              svreader->c_idx[3] = h;
            }
          }
        case SV_HEXAHEDRON:
          svwriter->write_hexahedron(svreader->c_idx, svreader->c_prop_f, c_final);
          break;
        case SV_ERROR:
          fprintf(stderr, "ERROR (sv2sv): svreader reported SV_ERROR\n.");
          exit(EXIT_FAILURE);
          break;
        default: 
          break;
        }
      }
    }
    else
    {
      while ((event = svreader->read_element())!=SV_EOF)
      {
        switch (event)
        {
        case SV_VERTEX:
          svwriter->write_vertex(svreader->v_prop_f);
          break;
        case SV_TETRAHEDRON:
          if (flip)
          {
            if (!reorient || !positive_orientation(svreaddereferenced->c_v_prop_f))
            {
              I32 h = svreader->c_idx[2];
              svreader->c_idx[2] = svreader->c_idx[3];
              svreader->c_idx[3] = h;
              bool b = svreader->c_final[2];
              svreader->c_final[2] = svreader->c_final[3];
              svreader->c_final[3] = b;
            }
          }
          svwriter->write_tetrahedron(svreader->c_idx, svreader->c_prop_f, svreader->c_final);
          break;
        case SV_HEXAHEDRON:
          svwriter->write_hexahedron(svreader->c_idx, svreader->c_prop_f, svreader->c_final);
          break;
        case SV_FINALIZED:
          svwriter->write_finalized(svreader->final_idx);
          break;
        case SV_ERROR:
          fprintf(stderr, "ERROR (sv2sv): svreader reported SV_ERROR\n.");
          exit(EXIT_FAILURE);
          break;
        default: 
          break;
        }
      }
    }
    svwriter->close(true, true);
    delete svwriter;
    delete svwriteopener;
  }
  else
  {
    fprintf(stderr,"read pass only.\n");
    
    svreader->SVreader::compute_bounding_box();

    fprintf(stderr,"# nverts %d\n",svreader->v_count);
    fprintf(stderr,"# ncells %d\n",svreader->c_count);
    fprintf(stderr,"# v_pnum_f %d\n",svreader->v_pnum_f);
    if (svreader->v_pnum_f && svreader->v_pmin_f)
    {
      fprintf(stderr,"# v_pmin_f");
      for (i = 0; i < svreader->v_pnum_f; i++) fprintf(stderr," %.8g",svreader->v_pmin_f[i]);
      fprintf(stderr,"\n");      
    }
    if (svreader->v_pnum_f && svreader->v_pmax_f)
    {
      fprintf(stderr,"# v_pmax_f");
      for (i = 0; i < svreader->v_pnum_f; i++) fprintf(stderr," %.8g",svreader->v_pmax_f[i]);
      fprintf(stderr,"\n");      
    }
    fprintf(stderr,"# c_pnum_f %d\n",svreader->c_pnum_f);
    if (svreader->c_pnum_f && svreader->c_pmin_f)
    {
      fprintf(stderr,"# c_pmin_f");
      for (i = 0; i < svreader->c_pnum_f; i++) fprintf(stderr," %.8g",svreader->c_pmin_f[i]);
      fprintf(stderr,"\n");      
    }
    if (svreader->c_pnum_f && svreader->c_pmax_f)
    {
      fprintf(stderr,"# c_pmax_f");
      for (i = 0; i < svreader->c_pnum_f; i++) fprintf(stderr," %.8g",svreader->c_pmax_f[i]);
      fprintf(stderr,"\n");      
    }
  }

  // stop clock
  clock_finish = clock();
  time_finish = time(0);
  fprintf(stderr,"needed %g ms/ %d seconds\n", 1000*(double)(clock_finish-clock_start)/CLOCKS_PER_SEC, (int)(time_finish-time_start));  

  svreader->close(true);
  delete svreader;
  delete svreadopener;

  return 0;
}
